<?php

namespace app\models;

use Yii;
use yii\db\ActiveRecord;
use yii\db\Query;
use yii\web\TooManyRequestsHttpException;

/**
 * @property $id
 * @property $auth_key
 * @property $deleted
 * @property $requests_limit - разрешенное кол-во запросов в минуту
 */
class UserModel extends ActiveRecord
{

    private $currentRequestId = 0;

    public static function tableName()
    {
        return 'users';
    }

    public function init()
    {
        parent::init();
        $this->on(self::EVENT_BEFORE_INSERT, function(){
            $this->deleted = 0;
            // по-умолчанию - 30 запросов в минуту (раз в 2 секунды)
            $this->requests_limit = 30;
        });
        $this->on(self::EVENT_AFTER_FIND, function(){

            // проверяем RateLimit
            if ($this->requests_limit > 0) {
                $requestsCount = $this->getRequests()->andWhere(['>', 'timestamp', time() - 30])->count();
                if($requestsCount > $this->requests_limit) {
                    throw new TooManyRequestsHttpException('Too many requests per minute');
                }
            }

            $request = new RequestModel();
            $request->user_id = $this->id;
            $request->timestamp = time();
            $request->save();
            $this->currentRequestId = $request->id;

        });
    }

    public function getRequests()
    {
        return $this->hasMany(RequestModel::className(), ['user_id' => 'id']);
    }

    public function generateAuthKey()
    {
        $this->auth_key = Yii::$app->security->generateRandomString(64);
    }

    public function delete()
    {
        $this->deleted = 1;
        $this->save(false);
    }

    public function getLastRequestTimestamp()
    {
        $requests = $this->getRequests();
        // не учитываем текущий запрос
        $requests->andWhere(['<>', 'id', $this->currentRequestId]);
        $requests->orderBy('id DESC');
        $lastRequest = $requests->limit(1)->one();
        if($lastRequest) {
            return $lastRequest->timestamp;
        } else {
            return 0;
        }
    }

    public function getDataModels($keys)
    {
        return DataModel::find()->where(['user_id' => $this->id, 'key_name' => $keys])->all();
    }

    public function setData(array $data)
    {
        // достаем и меняем ИМЕЮЩИЕСЯ ключи
        $keys = array_keys($data);
        $models = $this->getDataModels($keys);
        foreach ($models as $model) {
            /* @var $model DataModel */
            $key = $model->key_name;
            $model->value = $data[$key];
            $model->save();
            unset($data[$key]);
        }
        // добавляем НОВЫЕ ключи-значения
        if ($data) {
            foreach ($data as $key => $value) {
                $model = new DataModel();
                $model->user_id = $this->id;
                $model->key_name = $key;
                $model->value = $value;
                $model->save();
            }
        }
    }

    public function getData(array $keys)
    {
        $models = $this->getDataModels($keys);
        $result = [];
        $existKeys = [];
        foreach ($models as $model) {
            /* @var $model DataModel */
            $key = $model->key_name;
            $value = $model->value;
            $result[$key] = $value;
            $existKeys[] = $key;
        }
        $notExistKeys = array_diff($keys, $existKeys);
        if ($notExistKeys) {
            foreach ($notExistKeys as $key) {
                $result[$key] = null;
            }
        }
        return $result;
    }

    // общее количество дней, когда пользователь запускал приложение
    public function getRequestDaysCount()
    {
        // кешируем результат на сутки, поскольку запрос за "сегодня" уже учитывается
        // и нед неободимости пересчитывать результаты в течении этого времени.
        return Yii::$app->cache->getOrSet('requests_days', function() {
            return (new Query())
                ->select(['DATE(FROM_UNIXTIME(timestamp)) as day'])
                ->from(RequestModel::tableName())
                ->where(['user_id' => $this->id])
                ->groupBy('day')
                ->count();
        }, 3600 * 24);

    }

    public function getRequestDaysPeriodCount()
    {
        // кешируем результат на сутки, поскольку запрос за "сегодня" уже учитывается
        // и нед неободимости пересчитывать результаты в течении этого времени.
        return Yii::$app->cache->getOrSet('requests_periods', function(){
            // получаем дни, отсортированные в порядке возрастания, когда были запросы к API
            $days = (new Query())
                ->select(['DATE(FROM_UNIXTIME(timestamp)) as day'])
                ->from(RequestModel::tableName())
                ->where(['user_id' => $this->id])
                ->groupBy('day')
                ->orderBy('day ASC')
                ->all();

            if ($days) {
                $result = 0;
                $count = count($days);
                if ($count > 1) {
                    // todo: переименовать переменную во что-то понятное
                    $flag = false;
                    for ($i = 0; $i < $count - 1; ++$i) {
                        $day1 = strtotime($days[$i]['day']);
                        $day2 = strtotime($days[$i + 1]['day']);
                        // сравниваем, являются ли даты соседними
                        // так же имеем ввиду, что даты сравниваются по-парно,
                        // и при успешной N-1 проверке и неудачной N проверке, в результат нужно засчитвать сразу два дня.
                        if($day2 - $day1 <= 3600 * 12) {
                            ++$result;
                            $flag = true;
                        } elseif ($flag) {
                            ++$result;
                            $flag = false;
                        }
                    }
                    if ($flag) {
                        ++$result;
                    }
                    return $result;
                } else {
                    return 0;
                }
            } else {
                return 0;
            }
        }, 3600 * 24);
    }


}